/*
 * File: nejo.c
 *
 * (c) Peter Kleiweg
 *     Tue May 13 20:40:45 2008
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2,
 * or (at your option) any later version.
 *
 */

#define my_VERSION "0.01"

#define __NO_MATH_INLINES

#ifdef __WIN32__
#  define my_PATH_SEP '\\'
#else
#  define my_PATH_SEP '/'
#endif

#include <ctype.h>
#include <errno.h>
#include <math.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#ifdef __WIN32__
#  include <sys/timeb.h>
#endif

#define BUFSIZE 2048

int
    seed = 0,
    n_noise = 0,
    max_noise = 0,
    n_runs = 1,
    **connect,
    cophenetic = 0,
    arg_c,
    input_line = 0,
    size;

char
    *outfile = NULL,
    **labels,
    buffer [BUFSIZE + 1],
    **arg_v,
    *programname,
    *no_mem_buffer,
    out_of_memory [] = "Out of memory";

float
    *noise = NULL,
    **diff_in;
    
double
    **connect_d,
    **diff_cp;

void
    traverse (int start, int current, int prev, double sum),
    nj (double *D, int *N, int *edge1, int *edge2, double *edge_length),
    get_programname (char const *argv0),
    process_args (void),
    errit (char const *format, ...),
    syntax (void),
    *s_malloc (size_t size),
    *s_realloc (void *block, size_t size),
    checkCancel (void);
char
    *get_arg (void),
    *s_strdup (char const *s);
int
    getline (FILE *fp, int required, char const *filename);

int main (int argc, char *argv [])
{
    int
	i,
	j,
	k,
	N,
	run,
	nn_noise,
	runnoise,
	nodes = 0,
	*edge1,
	*edge2;
    float
	d,
	sd,
	sum,
	ssum;
    double
	*X,
	*edge_length;
    char
        *infile = NULL;
    FILE
        *fp = NULL;
    time_t
        tp;
#ifdef __WIN32__
    struct timeb
        tb;
#endif

    no_mem_buffer = (char *) malloc (1024);

    get_programname (argv [0]);

    arg_c = argc;
    arg_v = argv;
    process_args ();

    switch (arg_c) {
        case 1:
            if (isatty (fileno (stdin)))
                syntax ();
            fp = stdin;
            infile = "<stdin>";
            break;
        case 2:
            infile = arg_v [1];
            fp = fopen (infile, "r");
            if (! fp)
                errit ("Opening file \"%s\": %s", infile, strerror (errno));
            break;
        default:
            syntax ();
    }

    if (n_runs > 1 && ((! cophenetic) || ! n_noise))
	errit ("Using -r only makes sense in combination with -N and -c");

    getline (fp, 1, infile);
    if (sscanf (buffer, "%i", &size) != 1)
        errit (
            "file \"%s\", line %i\nTable size expected",
            infile,
            input_line
        );
    labels = (char **) s_malloc (size * sizeof (char *));
    for (i = 0; i < size; i++) {
        getline (fp, 1, infile);
        labels [i] = s_strdup (buffer);
    }

    diff_in = (float **) s_malloc (size * sizeof (float *));
    for (i = 1; i < size; i++)
        diff_in [i] = (float *) s_malloc (i * sizeof (float));
    for (i = 1; i < size; i++) {
        for (j = 0; j < i; j++) {
            getline (fp, 1, infile);
            if (sscanf (buffer, "%f", &d) != 1)
                errit ("file \"%s\", line %i\nValue expected", infile, input_line);
            diff_in [i][j] = d;
        }
    }

    if (fp != stdin)
        fclose (fp);

    sd = 0;
    if (n_noise) {

#ifdef __WIN32__
        ftime (&tb);
        srand (tb.millitm ^ (tb.time << 8));
#else
        srand (time (NULL) ^ (getpid () << 8));
#endif

        if (seed)
            srand (seed);

        k = 0;
        sum = ssum = 0.0;
        for (i = 0; i < size; i++)
            for (j = 0; j < i; j++) {
                sum += diff_in [i][j];
                ssum += diff_in [i][j] * diff_in [i][j];
                k++;
            }
        sd = sqrt ((ssum - sum * sum / (float) k) / (float) (k - 1));

    }
    nn_noise = n_noise ? n_noise : 1;

    if (cophenetic) {
	diff_cp = (double **) s_malloc (size * sizeof (double *));
	for (i = 1; i < size; i++) {
	    diff_cp [i] = (double *) s_malloc (i * sizeof (double));
	    for (j = 0; j < i; j++)
		diff_cp [i][j] = 0.0;
	}

	nodes = 2 * size - 2;
	connect = (int **) s_malloc (nodes * sizeof (int *));
	connect_d = (double **) s_malloc (nodes * sizeof (double *));
	for (i = 0; i < nodes; i++) {
	    connect [i] = (int *) s_malloc (3 * sizeof (int));
	    connect_d [i] = (double *) s_malloc (3 * sizeof (double));
	}
    }

    X = (double *) s_malloc ((size * size - size) / 2 * sizeof (double));

    edge1 = (int *) s_malloc((2*size - 3) * sizeof (int));
    edge2 = (int *) s_malloc((2*size - 3) * sizeof (int));
    edge_length = (double *) s_malloc((2*size - 3) * sizeof (double));

    for (run = 0; run < n_runs; run++) {

        for (runnoise = 0; runnoise < nn_noise; runnoise++) {

            checkCancel();

	    k = 0;
	    for (i = 0; i < size; i++)
		for (j = i + 1; j < size; j++)
		    X [k++] = diff_in [j][i];

	    N = size;

            if (n_noise)
                for (i = 0; i < k; i++)
		    X [i] += noise [runnoise] * sd * (((float) rand ()) / (float) RAND_MAX);

	    nj (X, &N, edge1, edge2, edge_length);

	    if (cophenetic) {
		for (i = 0; i < 2 * size - 3; i++) {
		    edge1 [i]--;
		    edge2 [i]--;
		}

		for (i = 0; i < nodes; i++)
		    connect [i][0] = connect [i][1] = connect [i][2] = -1;
		for (i = 0; i < 2*size-3; i++) {
		    j = edge1 [i];
		    k = edge2 [i];

		    if (connect [j][0] < 0) {
			connect [j][0] = k;
			connect_d [j][0] = edge_length [i];
		    } else if (connect [j][1] < 0) {
			connect [j][1] = k;
			connect_d [j][1] = edge_length [i];
		    } else {
			connect [j][2] = k;
			connect_d [j][2] = edge_length [i];
		    }

		    if (connect [k][0] < 0) {
			connect [k][0] = j;
			connect_d [k][0] = edge_length [i];
		    } else if (connect [k][1] < 0) {
			connect [k][1] = j;
			connect_d [k][1] = edge_length [i];
		    } else {
			connect [k][2] = j;
			connect_d [k][2] = edge_length [i];
		    }
		}

		for (i = 0; i < nodes; i++)
		    if (edge2 [i] < size)
			traverse (edge2 [i], edge2 [i], -1, 0);
	    }
	}
    }

    if (outfile) {
        fp = fopen (outfile, "w");
        if (! fp)
            errit ("Creating file \"%s\": %s", outfile, strerror (errno));
    } else
        fp = stdout;

    time (&tp);
    fprintf (
        fp,
        "# Created by %s, (c) Peter Kleiweg 2008\n"
	"# Built on nj.c, (c) Emmanuel Paradis 2006\n"
        "# More info: http://www.let.rug.nl/~kleiweg/L04/\n"
        "# Input file: %s\n",
        programname,
        infile
    );
    if (cophenetic) 
        fprintf (fp, "# Cophenetic differences\n");
    if (n_noise) {
        fputs ("# Noise:", fp);
        for (i = 0; i < n_noise; i++) {
            if (i && (i % 8 == 0))
                fprintf (fp, "\n#    ");
            fprintf (fp, " %g", noise [i]);
        }
        fputs ("\n", fp);
    }
    if (n_runs > 1)
        fprintf (fp, "# Runs: %i\n", n_runs);
    fprintf (
        fp,
        "# Date: %s\n",
        asctime (localtime (&tp))
    );

    if (cophenetic) {
	fprintf (fp, "%i\n", size);
	for (i = 0; i < size; i++)
	    fprintf (fp, "%s\n", labels [i]);
	for (i = 0; i < size; i++)
	    for (j = 0; j < i; j++)
		fprintf (fp, "%g\n", diff_cp [i][j] / n_runs / nn_noise);
    } else {
	fprintf (fp, 
	    ":: %i %g %g %g\n",
	    edge1 [2 * (size - 3)],
	    edge_length [2 * (size - 3)],
	    edge_length [2 * (size - 3) + 1],
	    edge_length [2 * (size - 3) + 2]
	);
	if (edge2 [2 * (size - 3)] <= size)
	    fprintf (fp, "L %s\n", labels [edge2 [2 * (size - 3)] - 1]);
	else
	    fprintf (fp, "C %i\n", edge2 [2 * (size - 3)]);
	if (edge2 [2 * (size - 3) + 1] <= size)
	    fprintf (fp, "L %s\n", labels [edge2 [2 * (size - 3) + 1] - 1]);
	else
	    fprintf (fp, "C %i\n", edge2 [2 * (size - 3) + 1]);
	if (edge2 [2 * (size - 3) + 2] <= size)
	    fprintf (fp, "L %s\n", labels [edge2 [2 * (size - 3) + 2] - 1]);
	else
	    fprintf (fp, "C %i\n", edge2 [2 * (size - 3) + 2]);
	for (i = 0; i < size - 3; i++) {
	    fprintf (fp, 
	        ": %i %g %g\n",
		edge1 [2 * i],
		edge_length [2 * i],
		edge_length [2 * i + 1]
	    );
	    if (edge2 [2 * i] <= size)
		fprintf (fp, "L %s\n", labels [edge2 [2 * i] - 1]);
	    else
		fprintf (fp, "C %i\n", edge2 [2 * i]);
	    if (edge2 [2 * i + 1] <= size)
		fprintf (fp, "L %s\n", labels [edge2 [2 * i + 1] - 1]);
	    else
		fprintf (fp, "C %i\n", edge2 [2 * i + 1]);
	}	
    }

    if (outfile)
        fclose (fp);

    return 0;
}

void traverse (int start, int current, int prev, double sum)
{
    int
	next,
	i;

    if (current != start && current < size) {
	if (start > current)
	    diff_cp [start][current] += sum / 2.0;
	else
	    diff_cp [current][start] += sum / 2.0;
	return;
    }

    for (i = 0; i < 3; i++) {
	next = connect [current][i];
	if (next >= 0 && next != prev)
	    traverse (start, next, current, sum + connect_d [current][i]);
    }
}

void checkCancel ()
{
    if (access ("_CANCEL_.L04", F_OK) == 0)
        errit ("CANCELLED");
}

int getline (FILE *fp, int required, char const *filename)
/* Lees een regel in
 * Plaats in buffer
 * Negeer lege regels en regels die beginnen met #
 * Verwijder leading/trailing white space
 */
{
    int
        i;

    for (;;) {
        if (fgets (buffer, BUFSIZE, fp) == NULL) {
            if (required)
                errit ("Unexpected end of file in \"%s\"", filename);
            else
                return 0;
        }
        input_line++;
        i = strlen (buffer);
        while (i && isspace ((unsigned char) buffer [i - 1]))
            buffer [--i] = '\0';
        i = 0;
        while (buffer [i] && isspace ((unsigned char) buffer [i]))
            i++;
        if (buffer [i] == '#')
            continue;
        if (buffer [i]) {
            memmove (buffer, buffer + i, strlen (buffer + i) + 1);
            return 1;
        }
    }
}

void process_args ()
{
    float
	f;

    while (arg_c > 1 && arg_v [1][0] == '-') {
        switch (arg_v [1][1]) {
            case 'c':
		cophenetic = 1;
		break;
	    case 'N':
		f = atof (get_arg ());
		if (f < 0.0)
		    errit ("Illegal value for -N: %g", f);
		if (n_noise == max_noise) {
		    max_noise += 64;
		    noise = (float *) s_realloc (noise, max_noise * sizeof (float));
		}
		noise [n_noise++] = f;
		break;
            case 'o':
                outfile = get_arg ();
                break;
            case 'r':
                n_runs = atoi (get_arg ());
                break;
            case 's':
		seed = atoi (get_arg ());
		if (seed < 1)
		    errit ("Option -s : seed must be positive");
                break;
            default:
                errit ("Illegal option '%s'", arg_v [1]);
        }
	arg_c--;
	arg_v++;
    }
}

char *get_arg ()
{
    if (arg_v [1][2])
        return arg_v [1] + 2;

    if (arg_c == 2)
        errit ("Missing argument for '%s'", arg_v [1]);

    arg_v++;
    arg_c--;
    return arg_v [1];
}

void errit (char const *format, ...)
{
    va_list
	list;

    fprintf (stderr, "\nError %s: ", programname);

    va_start (list, format);
    vfprintf (stderr, format, list);

    fprintf (stderr, "\n\n");

    exit (1);
}

void get_programname (char const *argv0)
{
#ifdef __MSDOS__
    char
        name [MAXFILE];
    fnsplit (argv0, NULL, NULL, name, NULL);
    programname = strdup (name);
#else
    char
        *p;
    p = strrchr (argv0, my_PATH_SEP);
    if (p)
        programname = strdup (p + 1);
    else
        programname = strdup (argv0);
#endif
}

void *s_malloc (size_t size)
{
    void
	*p;

    p = malloc (size);
    if (! p) {
        free (no_mem_buffer);
	errit (out_of_memory);
    }
    return p;
}

void *s_realloc (void *block, size_t size)
{
    void
        *p;

    p = realloc (block, size);
    if (! p) {
        free (no_mem_buffer);
        errit (out_of_memory);
    }
    return p;
}

char *s_strdup (char const *s)
{
    char
	*s1;

    if (s) {
	s1 = (char *) s_malloc (strlen (s) + 1);
	strcpy (s1, s);
    } else {
	s1 = (char *) s_malloc (1);
	s1 [0] = '\0';
    }
    return s1;
}

void syntax ()
{
    fprintf (
	stderr,
	"\n"
	"Neighbor Joining Clustering, Version " my_VERSION "\n"
	"(c) Peter Kleiweg 2008\n"
	"Built on nj.c, (c) Emmanuel Paradis 2006\n"
	"\n"
	"Usage: %s [-c] [-N float] [-o filename] [-r int] [-s int]\n"
	"\t    [difference table file]\n"
	"\n"
	"\t-c : cophenetic difference output instead of tree\n"
	"\t-N : noise level\n"
	"\t-o : output file\n"
	"\t-r : number of runs (with -c and -N)\n"
	"\t-s : seed for random number generator\n"
	"\n"
	"Option -N can be used more than once\n"
	"\n",
	programname
    );
    exit (1);
}

/****************************************************************/

/* nj.c       2006-11-13 */

/* Copyright 2006 Emmanuel Paradis */

/* This file is part of the R-package `ape'. */
/* See the file ../COPYING for licensing issues. */

#define DINDEX(i, j) n*(i - 1) - i*(i - 1)/2 + j - i - 1

int give_index(int i, int j, int n)
{
    if (i > j) return(DINDEX(j, i));
    else return(DINDEX(i, j));
}

double sum_dist_to_i(int n, double *D, int i)
/* returns the sum of all distances D_ij between i and j
   with j between 1, and n and j != i */
{
    double sum;
    int j;

    sum = 0;

    if (i != 1) {
        for (j = 1; j < i; j++)
	  sum += D[DINDEX(j, i)];
    }

    if (i != n) {
        for (j = i + 1; j <= n; j++)
	  sum += D[DINDEX(i, j)];
    }

    return(sum);
}

#define GET_I_AND_J                                               \
/* Find the 'R' indices of the two corresponding OTUs */          \
/* The indices of the first element of the pair in the            \
   distance matrix are n-1 times 1, n-2 times 2, n-3 times 3,     \
   ..., once n-1. Given this, the algorithm below is quite        \
   straightforward.*/                                             \
    i = 0;                                                        \
    for (OTU1 = 1; OTU1 < n; OTU1++) {                            \
        i += n - OTU1;                                            \
	if (i >= smallest + 1) break;                             \
    }                                                             \
    /* Finding the second OTU is easier! */                       \
    OTU2 = smallest + 1 + OTU1 - n*(OTU1 - 1) + OTU1*(OTU1 - 1)/2;

#define SET_CLADE                           \
/* give the node and tip numbers to edge */ \
    edge2[k] = otu_label[OTU1 - 1];         \
    edge2[k + 1] = otu_label[OTU2 - 1];     \
    edge1[k] = edge1[k + 1] = cur_nod;

void nj(double *D, int *N, int *edge1, int *edge2, double *edge_length)
{
    double SUMD, Sdist, *S, Ndist, *new_dist, A, B, *DI, d_i, x, y;
    int n, i, j, k, ij, smallest, OTU1, OTU2, cur_nod, o_l, *otu_label;

    S = &Sdist;
    new_dist = &Ndist;
    otu_label = &o_l;
    DI = &d_i;

    n = *N;
    cur_nod = 2*n - 2;

    S = (double*)s_malloc(n*(n - 1)/2 * sizeof(double));
    new_dist = (double*)s_malloc(n*(n - 1)/2 * sizeof(double));
    otu_label = (int*)s_malloc(n * sizeof(int));
    DI = (double*)s_malloc((n - 2) * sizeof(double));

    for (i = 0; i < n; i++) otu_label[i] = i + 1;
    k = 0;

    /* First, look if there are distances equal to 0. */
    /* Since there may be multichotomies, we loop
       through the OTUs instead of the distances. */

    OTU1 = 1;
    while (OTU1 < n) {
        OTU2 = OTU1 + 1;
        while (OTU2 <= n) {
	    if (D[DINDEX(OTU1, OTU2)] == 0) {
		SET_CLADE
		edge_length[k] = edge_length[k + 1] = 0.;
		k = k + 2;

		/* update */
		/* We remove the second tip label: */
		if (OTU2 < n) {
  		    for (i = OTU2; i < n; i++)
		      otu_label[i - 1] = otu_label[i];
		}

		ij = 0;
		for (i = 1; i < n; i++) {
		    if (i == OTU2) continue;
		    for (j = i + 1; j <= n; j++) {
		        if (j == OTU2) continue;
			new_dist[ij] = D[DINDEX(i, j)];
			ij++;
		    }
		}
		n--;
		for (i = 0; i < n*(n - 1)/2; i++) D[i] = new_dist[i];

		otu_label[OTU1 - 1] = cur_nod;
		/* to avoid adjusting the internal branch at the end: */
		DI[cur_nod - *N - 1] = 0;
		cur_nod--;
	    } else OTU2++;
	}
	OTU1++;
    }

    while (n > 3) {

        SUMD = 0;
	for (i = 0; i < n*(n - 1)/2; i++) SUMD += D[i];

	for (i = 1; i < n; i++) {
	    for (j = i + 1; j <= n; j++) {
	        /* we know that i < j, so: */
	        ij =  DINDEX(i, j);
	        A = sum_dist_to_i(n, D, i) - D[ij];
	        B = sum_dist_to_i(n, D, j) - D[ij];
	        S[ij] = (A + B)/(2*n - 4) + 0.5*D[ij] + (SUMD - A - B - D[ij])/(n - 2);
	    }
	}

	/* find the 'C' index of the smallest value of S */
	smallest = 0;
	for (i = 1; i < n*(n - 1)/2; i++)
	  if (S[smallest] > S[i]) smallest = i;

	GET_I_AND_J

	SET_CLADE

        /* get the distances between all OTUs but the 2 selected ones
           and the latter:
             a) get the sum for both
	     b) compute the distances for the new OTU */
        A = B = ij = 0;
        for (i = 1; i <= n; i++) {
            if (i == OTU1 || i == OTU2) continue;
            x = D[give_index(i, OTU1, n)]; /* distance between OTU1 and i */
            y = D[give_index(i, OTU2, n)]; /* distance between OTU2 and i */
            new_dist[ij] = (x + y)/2;
            A += x;
            B += y;
            ij++;
        }
        /* compute the branch lengths */
        A /= n - 2;
        B /= n - 2;
        edge_length[k] = (D[smallest] + A - B)/2;
        edge_length[k + 1] = (D[smallest] + B - A)/2;
        DI[cur_nod - *N - 1] = D[smallest];

        /* update before the next loop */
        if (OTU1 > OTU2) { /* make sure that OTU1 < OTU2 */
            i = OTU1;
	    OTU1 = OTU2;
	    OTU2 = i;
        }
        if (OTU1 != 1)
          for (i = OTU1 - 1; i > 0; i--) otu_label[i] = otu_label[i - 1];
        if (OTU2 != n)
          for (i = OTU2; i <= n; i++) otu_label[i - 1] = otu_label[i];
        otu_label[0] = cur_nod;

        for (i = 1; i < n; i++) {
            if (i == OTU1 || i == OTU2) continue;
	    for (j = i + 1; j <= n; j++) {
	        if (j == OTU1 || j == OTU2) continue;
		new_dist[ij] = D[DINDEX(i, j)];
		ij++;
	    }
        }

	n--;
	for (i = 0; i < n*(n - 1)/2; i++) D[i] = new_dist[i];

	cur_nod--;
	k = k + 2;
    }

    for (i = 0; i < 3; i++) {
        edge1[*N*2 - 4 - i] = cur_nod;
	edge2[*N*2 - 4 - i] = otu_label[i];
    }

    edge_length[*N*2 - 4] = (D[0] + D[1] - D[2])/2;
    edge_length[*N*2 - 5] = (D[0] + D[2] - D[1])/2;
    edge_length[*N*2 - 6] = (D[2] + D[1] - D[0])/2;

    for (i = 0; i < *N*2 - 3; i++) {
        if (edge2[i] <= *N) continue;
	/* In case there are zero branch lengths (see above): */
	if (DI[edge2[i] - *N - 1] == 0) continue;
	edge_length[i] -= DI[edge2[i] - *N - 1]/2;
    }

    free (DI);
    free (otu_label);
    free (new_dist);
    free (S);
}
